-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- RadioHDL wrapper

LIBRARY ieee, technology_lib;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
USE technology_lib.technology_pkg.all;

LIBRARY altera_lnsim;
USE altera_lnsim.altera_lnsim_components.all;

ENTITY ip_arria10_e3sge3_ram_r_w IS
  GENERIC (
    g_inferred   : BOOLEAN := FALSE;
    g_adr_w      : NATURAL := 5;
    g_dat_w      : NATURAL := 8;
    g_nof_words  : NATURAL := 2**5;
    g_rd_latency : NATURAL := 1;     -- choose 1 or 2
    g_init_file  : STRING  := "UNUSED"
  );
  PORT (
    clk         : IN STD_LOGIC  := '1';
    data        : IN STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0) := (OTHERS=>'0');
    rdaddress   : IN STD_LOGIC_VECTOR(g_adr_w-1 DOWNTO 0) := (OTHERS=>'0');
    wraddress   : IN STD_LOGIC_VECTOR(g_adr_w-1 DOWNTO 0) := (OTHERS=>'0');
    wren        : IN STD_LOGIC  := '0';
    q           : OUT STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0)
  );
END ip_arria10_e3sge3_ram_r_w;


ARCHITECTURE SYN OF ip_arria10_e3sge3_ram_r_w IS

  CONSTANT c_outdata_reg_b : STRING := tech_sel_a_b(g_rd_latency=1, "UNREGISTERED", "CLOCK1");

  COMPONENT altera_syncram
  GENERIC (
          address_aclr_b  : string;
          address_reg_b  : string;
          clock_enable_input_a  : string;
          clock_enable_input_b  : string;
          clock_enable_output_b  : string;
          init_file  : string;
          intended_device_family  : string;
          lpm_type  : string;
          numwords_a  : integer;
          numwords_b  : integer;
          operation_mode  : string;
          outdata_aclr_b  : string;
          outdata_reg_b  : string;
          power_up_uninitialized  : string;
          widthad_a  : integer;
          widthad_b  : integer;
          width_a  : integer;
          width_b  : integer;
          width_byteena_a  : integer
  );
  PORT (
      address_a : in std_logic_vector(g_adr_w-1 downto 0);
      address_b : in std_logic_vector(g_adr_w-1 downto 0);
      clock0 : in std_logic;
      data_a : in std_logic_vector(g_dat_w-1 downto 0);
      wren_a : in std_logic;
      q_b : out std_logic_vector(g_dat_w-1 downto 0)
  );
  END COMPONENT;

  SIGNAL rdaddr : natural range 0 to g_nof_words - 1;
  SIGNAL wraddr : natural range 0 to g_nof_words - 1;
  
  SIGNAL out_q  : STD_LOGIC_VECTOR (g_dat_w-1 DOWNTO 0);
  SIGNAL reg_q  : STD_LOGIC_VECTOR (g_dat_w-1 DOWNTO 0);
  
BEGIN

  ASSERT g_rd_latency=1 OR g_rd_latency=2 REPORT "ip_arria10_e3sge3_ram_r_w : read latency must be 1 (default) or 2" SEVERITY FAILURE;
  
  gen_ip : IF g_inferred=FALSE GENERATE
    -- Copied from ip_arria10_e3sge3_ram_r_w/ram_2port_140/sim/ip_arria10_e3sge3_ram_r_w_ram_2port_140_hukd7xi.vhd
    u_altera_syncram : altera_syncram
    GENERIC MAP (
            address_aclr_b  => "NONE",
            address_reg_b  => "CLOCK0",
            clock_enable_input_a  => "BYPASS",
            clock_enable_input_b  => "BYPASS",
            clock_enable_output_b  => "BYPASS",
            init_file  => g_init_file,
            intended_device_family  => "Arria 10",
            lpm_type  => "altera_syncram",
            numwords_a  => g_nof_words,
            numwords_b  => g_nof_words,
            operation_mode  => "DUAL_PORT",
            outdata_aclr_b  => "NONE",
            outdata_reg_b  => c_outdata_reg_b,
            power_up_uninitialized  => "FALSE",
            widthad_a  => g_adr_w,
            widthad_b  => g_adr_w,
            width_a  => g_dat_w,
            width_b  => g_dat_w,
            width_byteena_a  => 1
    )
    PORT MAP (
        address_a => wraddress,
        address_b => rdaddress,
        clock0 => clk,
        data_a => data,
        wren_a => wren,
        q_b => q
    );
  END GENERATE;
  
  gen_inferred : IF g_inferred=TRUE GENERATE
    rdaddr <= TO_INTEGER(UNSIGNED(rdaddress));
    wraddr <= TO_INTEGER(UNSIGNED(wraddress));
  
    u_mem : entity work.ip_arria10_e3sge3_simple_dual_port_ram_single_clock
    generic map (
      DATA_WIDTH => g_dat_w,
      ADDR_WIDTH => g_adr_w
    )
    port map (
      clk   => clk,
      raddr => rdaddr,
      waddr => wraddr,
      data  => data,
      we    => wren,
      q     => out_q
    );
  
    reg_q <= out_q WHEN rising_edge(clk);
    
    q <= out_q WHEN g_rd_latency=1 ELSE reg_q;
  END GENERATE;
   
END SYN;
