--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------


-- Author : J Hargreaves
-- Purpose: Technology selection wrapper to instantiate 
--   JESD204b interface for ADCs and DACs
-- Description:
--
--   Select g_direction = RX_ONLY for ADC
--   IP settings are for the AD9683 ADC
--   Technology currently supported: Intel Arria 10 (e1sg), unb2b, unb2c
--
-- . For c_tech_arria10_e1sg:
--
--     device_clk (ADC sample clock) -->
--     sysref (ADC 1PPS) -->
--                ___________________
--                |                 |
--     tx_snk --->|tech_jesd204b    |---> sync_n
--     rx_src <---|_arria10_e1sg    |<--- serial_rx
--                |_________________|
--                    |
--                    |
--                  mac_mm
--
--
-- ToDo: Change g_nof_channels to g_nof_streams in IP

LIBRARY IEEE, common_lib, dp_lib, technology_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE technology_lib.technology_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.tech_jesd204b_component_pkg.ALL;

ENTITY tech_jesd204b IS
  GENERIC (
    g_sim                 : BOOLEAN := FALSE;
    g_technology          : NATURAL := c_tech_select_default;
    g_nof_streams         : NATURAL := 12;
    g_nof_sync_n          : NATURAL := 12;
    g_direction           : STRING  := "RX_ONLY"; -- "TX_RX", "TX_ONLY", "RX_ONLY"
    g_jesd_freq           : STRING  := "200MHz"
  );
  PORT (
    -- JESD204B external signals
    jesd204b_refclk       : IN STD_LOGIC := '0';                             -- Reference clock. For AD9683 use 200MHz direct from clock reference pin
    jesd204b_sysref       : IN STD_LOGIC := '0';                             -- SYSREF should drive ADC and FPGA with correct phase wrt jesd204b_device_clk 
    jesd204b_sync_n_arr   : OUT STD_LOGIC_VECTOR(g_nof_sync_n-1 DOWNTO 0);   -- output to control ADC initialization/syncronization phase
    
    jesd204b_disable_arr  : IN STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);

    -- Data to fabric
    rx_sosi_arr           : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);    -- Parallel data out to fabric
    rx_clk                : OUT  STD_LOGIC := '0';                           -- Exported data clock (frame clock) to fabric
    rx_rst                : OUT  STD_LOGIC := '0';                           -- Exported reset on rx_clk domain
    rx_sysref             : OUT  STD_LOGIC := '0';                           -- Exported copy of sysref

    -- MM Control
    mm_clk                : IN  STD_LOGIC;
    mm_rst                : IN  STD_LOGIC;
         
    jesd204b_mosi         : IN  t_mem_mosi;         --  mm control
    jesd204b_miso         : OUT t_mem_miso; 
    
    -- Serial connections to transceiver pins
    serial_tx_arr         : OUT STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);      -- Not used for ADC
    serial_rx_arr         : IN  STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0)
  );
END tech_jesd204b;


ARCHITECTURE str OF tech_jesd204b IS
BEGIN
                   
  gen_ip_arria10_e1sg : IF g_technology=c_tech_arria10_e1sg GENERATE
    u0 : ENTITY work.tech_jesd204b_arria10_e1sg
    GENERIC MAP(
      g_sim                => g_sim,                
      g_nof_streams        => g_nof_streams,      
      g_nof_sync_n         => g_nof_sync_n,        
      g_direction          => g_direction,
      g_jesd_freq          => g_jesd_freq
    )
    PORT MAP(
      jesd204b_refclk      => jesd204b_refclk,   
      jesd204b_sysref      => jesd204b_sysref,   
      jesd204b_sync_n_arr  => jesd204b_sync_n_arr,   
  
      jesd204b_disable_arr => jesd204b_disable_arr,

      rx_src_out_arr       => rx_sosi_arr,          
      rx_clk               => rx_clk,          
      rx_rst               => rx_rst,          
      rx_sysref            => rx_sysref,          
  
      -- MM
      mm_clk               => mm_clk,           
      mm_rst               => mm_rst,           
  
      jesd204b_mosi        => jesd204b_mosi,         
      jesd204b_miso        => jesd204b_miso,         
  
       -- Serial
      serial_tx_arr        => serial_tx_arr,
      serial_rx_arr        => serial_rx_arr
    );
  END GENERATE;

  gen_ip_arria10_e2sg : IF g_technology=c_tech_arria10_e2sg GENERATE
    u0 : ENTITY work.tech_jesd204b_arria10_e2sg
    GENERIC MAP(
      g_sim                => g_sim,                
      g_nof_streams        => g_nof_streams,      
      g_nof_sync_n         => g_nof_sync_n,       
      g_direction          => g_direction
    )
    PORT MAP(
      jesd204b_refclk      => jesd204b_refclk,   
      jesd204b_sysref      => jesd204b_sysref,   
      jesd204b_sync_n_arr  => jesd204b_sync_n_arr,   
  
      rx_src_out_arr       => rx_sosi_arr,          
      rx_clk               => rx_clk,          
      rx_rst               => rx_rst,          
      rx_sysref            => rx_sysref,          
  
      -- MM
      mm_clk               => mm_clk,           
      mm_rst               => mm_rst,           
  
      jesd204b_mosi        => jesd204b_mosi,         
      jesd204b_miso        => jesd204b_miso,         
  
       -- Serial
      serial_tx_arr        => serial_tx_arr,
      serial_rx_arr        => serial_rx_arr
    );
  END GENERATE;

END str;
