-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Reusable test bench link connection for the tech_mac_10g
-- Description:
-- . Support XGMII layer connect
-- . Support PHY serial layer connect with optional link_fault
-- . Support link delay

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;


ENTITY tb_tech_mac_10g_link_connect IS
  GENERIC (
    g_loopback    : BOOLEAN := TRUE;  -- default TRUE for loopback tx to rx, else use FALSE to connect tx-tx, rx-rx between two tb devices
    g_link_delay  : TIME :=  0 ns
  );
  PORT (
    link_fault    : IN  STD_LOGIC := '0';  -- when '1' then forces rx_serial='0' and xaui_rx="0000"

    -- XGMII layer connect
    xgmii_tx_data : IN  STD_LOGIC_VECTOR(c_xgmii_w-1 DOWNTO 0) := (OTHERS=>'X');        -- 72 bit
    xgmii_rx_data : OUT STD_LOGIC_VECTOR(c_xgmii_w-1 DOWNTO 0);                         -- 72 bit
    
    -- 10GBASE-R serial layer connect
    serial_tx     : IN  STD_LOGIC := 'X';
    serial_rx     : OUT STD_LOGIC;                                                      -- connects to delayed serial_tx when g_loopback=TRUE else to serial_rx_in
    serial_tx_out : OUT STD_LOGIC;                                                      -- connects to delayed serial_tx
    serial_rx_in  : IN  STD_LOGIC := 'X';                                               -- used when g_loopback=FALSE
    
    -- XAUI serial layer connect
    xaui_tx       : IN  STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0) := (OTHERS=>'X');
    xaui_rx       : OUT STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0);                  -- connects to delayed xaui_tx when g_loopback=TRUE else to xaui_rx_in
    xaui_tx_out   : OUT STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0);                  -- connects to delayed xaui_tx
    xaui_rx_in    : IN  STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0) := (OTHERS=>'X')  -- used when g_loopback=FALSE
  );
END tb_tech_mac_10g_link_connect;


ARCHITECTURE tb OF tb_tech_mac_10g_link_connect IS
  
  SIGNAL serial_tx_dly   : STD_LOGIC;
  SIGNAL xaui_tx_dly     : STD_LOGIC_VECTOR(c_nof_xaui_lanes-1 DOWNTO 0);

BEGIN
  -----------------------------------------------------------------------------
  -- XGMII layer link
  -----------------------------------------------------------------------------
  xgmii_rx_data <= TRANSPORT xgmii_tx_data AFTER g_link_delay;

  -----------------------------------------------------------------------------
  -- 10GBASE-R serial layer link
  -----------------------------------------------------------------------------
  serial_tx_dly <= TRANSPORT serial_tx AFTER g_link_delay;
  serial_tx_out <= serial_tx_dly;
  
  p_serial_rx : PROCESS(serial_tx_dly, serial_rx_in, link_fault)
  BEGIN
    IF g_loopback=TRUE THEN
      serial_rx <= serial_tx_dly;
    ELSE
      serial_rx <= serial_rx_in;
    END IF;
    IF link_fault='1' THEN
      serial_rx <= '0';
    END IF;
  END PROCESS;

  -----------------------------------------------------------------------------
  -- XAUI serial layer link
  -----------------------------------------------------------------------------
  xaui_tx_dly <= TRANSPORT xaui_tx AFTER g_link_delay;
  xaui_tx_out <= xaui_tx_dly;
  
  p_xaui_rx : PROCESS(xaui_tx_dly, xaui_rx_in, link_fault)
  BEGIN
    IF g_loopback=TRUE THEN
      xaui_rx <= xaui_tx_dly;
    ELSE
      xaui_rx <= xaui_rx_in;
    END IF;
    IF link_fault='1' THEN
      xaui_rx <= (OTHERS=>'0');
    END IF;
  END PROCESS;

END tb;
