--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Author: Eric Kooistra / Daniel van der Schuur
-- Purpose:
--   Drop-in simulation model for tech_transceiver_gx.vhd.
-- Description:
--   A fast serdes for simulation within tr_nonbonded.
--
--   Internally the GX model divides the c_tr_clk_period up into g_data_w * 10/8
--   smaller periods that transport the data and the overhead. The model uses 
--   some artifical 10b/8b scheme to encode the overhead. The GX model is
--   inspired on 10/8 encoding, however it will also suit 66/64 encoding,
--   because the GX model does not model the encoding, it only ensures that the
--   data can be transported as a bit stream.
--
--   The c_tr_clk_period is set sufficiently short to fit within any actual
--   tr_clk period. This does not slow down the simulation, because
--   after the last data has been processed the GX model waits for the next 
--   tr_clk rising_edge, so the amount of toggling is independent of
--   c_tr_clk_period. To be sure the model measures the actual tr_clk period,
--   which is assumed to be related to the user data rate clock. The actual
--   tr_clk_period must be >= c_tr_clk_period, because otherwise the GX model
--   will miss data. After every data word the GX serializer and deserializer
--   models realign with the actual tr_clk. Therefore it is possible to have
--   tr_clk_period > c_tr_clk_period. 
--
--   The sim model can be stopped by external tb_end input or by stopping the
--   tr_clk, which is detected via internal tr_end. Using tr_end is useful
--   when tb_end cannot be provided via the tb, e.g. because tb_end is not
--   available as IN port on top level FPGA entities. Using tb_end IN is kept
--   to be remain compatible with tb that use it.
--
-- Remarks:
--   None

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_str_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

ENTITY sim_transceiver_gx IS 
  GENERIC(
    g_data_w     : NATURAL;
    g_nof_gx     : NATURAL;
    g_mbps       : NATURAL;  -- line rate in Mbps so including 10/8 encoding
    g_tx         : BOOLEAN;
    g_rx         : BOOLEAN
  );      
  PORT(
    tb_end       : IN  STD_LOGIC := '0';  -- in simulation stop internal clocks when tb_end='1' to support 'run -all'
    
    tr_clk       : IN  STD_LOGIC;

    rx_clk       : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    rx_rst       : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    rx_sosi_arr  : OUT t_dp_sosi_arr(g_nof_gx-1 DOWNTO 0);
    rx_siso_arr  : IN  t_dp_siso_arr(g_nof_gx-1 DOWNTO 0);

    tx_clk       : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    tx_rst       : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    tx_sosi_arr  : IN  t_dp_sosi_arr(g_nof_gx-1 DOWNTO 0);
    tx_siso_arr  : OUT t_dp_siso_arr(g_nof_gx-1 DOWNTO 0);
   
    --Serial I/O
    rx_datain    : IN  STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0);
    tx_dataout   : OUT STD_LOGIC_VECTOR(g_nof_gx-1 DOWNTO 0)
  );

END sim_transceiver_gx;

ARCHITECTURE str OF sim_transceiver_gx IS

  CONSTANT c_data_clk_period  : TIME := 1 us * g_data_w * 10 / 8 / g_mbps;
  CONSTANT c_tr_clk_period    : TIME := 1 ns;   -- run the GX model somewhat faster to support shorter tr_clk_period

  TYPE t_ctrl_2arr IS ARRAY(g_nof_gx-1 DOWNTO 0) OF STD_LOGIC_VECTOR(g_data_w/c_byte_w-1 DOWNTO 0);
  
  SIGNAL tr_rst      : STD_LOGIC;

  SIGNAL tx_ready    : STD_LOGIC;

  SIGNAL tx_in_ctrl  : t_ctrl_2arr;
  SIGNAL tx_in_sop   : t_ctrl_2arr;
  SIGNAL tx_in_eop   : t_ctrl_2arr;
  SIGNAL rx_out_ctrl : t_ctrl_2arr;
  SIGNAL rx_out_sop  : t_ctrl_2arr;
  SIGNAL rx_out_eop  : t_ctrl_2arr;
  
  SIGNAL dbg_c_data_clk_period : TIME := c_data_clk_period;  -- view constant in Wave window
  SIGNAL dbg_c_tr_clk_period   : TIME := c_tr_clk_period;    -- view constant in Wave window
  SIGNAL tr_clk_period         : TIME;
  
  SIGNAL tr_end      : STD_LOGIC := '0';  -- local detect whether the tr_clk has stopped by continuosly checking the toggling
  SIGNAL sim_end     : STD_LOGIC := '0';  -- in simulation stop internal clocks when tb_end='1' or tr_end='1' to support 'run -all'

BEGIN

  u_areset_tr_rst : ENTITY common_lib.common_areset
  GENERIC MAP(
    g_rst_level => '1'
  )
  PORT MAP(
    clk     => tr_clk,
    in_rst  => '0',
    out_rst => tr_rst
  );  

  p_check_tr_clk_period : PROCESS
    VARIABLE v_period : TIME;
  BEGIN
    -- wait until tr_rst released, so tr_clk is running
    WAIT UNTIL tr_rst='0';
    
    -- measure tr_clk period
    WAIT UNTIL rising_edge(tr_clk);
    v_period := NOW;
    WAIT UNTIL rising_edge(tr_clk);
    v_period := NOW - v_period;
    WAIT UNTIL rising_edge(tr_clk);
    ASSERT v_period>=c_tr_clk_period REPORT "Actual tr_clk period = " & time_to_str(v_period) & " should be >= c_tr_clk_period = " & time_to_str(c_tr_clk_period) SEVERITY ERROR;
    
    -- use signal to show tr_clk period in Wave window
    tr_clk_period <= v_period;
    
    -- detect whether tr_clk has stopped by continuously checking the toggling (cannot use WAIT UNTIL
    -- rising_edge(tr_clk) because tr_clk is the item that is checked).
    WAIT UNTIL rising_edge(tr_clk);
    WAIT FOR c_tr_clk_period/4;  -- to be sure that tr_clk='1' before starting the while TRUE loop
    WHILE TRUE LOOP
      IF tr_clk/='1' THEN  -- level of tr_clk after rising_edge()
        EXIT;
      END IF;
      WAIT FOR v_period/2;
      IF tr_clk/='0' THEN  -- level of tr_clk after half period
        EXIT;
      END IF;
      WAIT FOR v_period/2;
    END LOOP;
      
    tr_end <= '1';
    WAIT;
  END PROCESS;
  
  sim_end <= tb_end OR tr_end;  -- combine external and internal stop sim conditions
  
  tx_clk <= (OTHERS=>tr_clk);
  tx_rst <= (OTHERS=>tr_rst);
  rx_clk <= (OTHERS=>tr_clk);
  rx_rst <= (OTHERS=>tr_rst);

  p_tx_rdy : PROCESS
    VARIABLE v_tx_clk_cnt : NATURAL := 0;
  BEGIN
    tx_ready <= '0';
    WAIT FOR 1084.8 ns;              -- Time until tx_rst (in tr_clk domain) is deasserted: tx init done
    WHILE v_tx_clk_cnt < 1000 LOOP   -- 1000 tx clk cycles until alignment is completed
      WAIT UNTIL rising_edge(tr_clk);
      v_tx_clk_cnt := v_tx_clk_cnt + 1;
      tx_ready <= '1';
    END LOOP;
    WAIT;
  END PROCESS;

  gen_sim: FOR i IN 0 to g_nof_gx-1 GENERATE
    gen_tx : IF g_tx = TRUE GENERATE
      tx_siso_arr(i).ready <= tx_ready;
      tx_siso_arr(i).xon   <= tx_ready;

      tx_in_ctrl(i) <= (OTHERS=>tx_sosi_arr(i).valid);
      tx_in_sop(i)  <= (OTHERS=>tx_sosi_arr(i).sop);
      tx_in_eop(i)  <= (OTHERS=>tx_sosi_arr(i).eop);

      u_ser: ENTITY work.sim_transceiver_serializer
      GENERIC MAP (
        g_data_w         => g_data_w,
        g_tr_clk_period  => c_tr_clk_period
      )
      PORT MAP (
        tb_end        => sim_end,
        
        tr_clk        => tr_clk,  
        tr_rst        => tr_rst,
         
        tx_in_data    => tx_sosi_arr(i).data(g_data_w-1 DOWNTO 0),
        tx_in_ctrl    => tx_in_ctrl(i),
        tx_in_sop     => tx_in_sop(i),
        tx_in_eop     => tx_in_eop(i),
    
        tx_serial_out => tx_dataout(i)
      );
    END GENERATE;

    gen_rx : IF g_rx = TRUE GENERATE   
      u_des: ENTITY work.sim_transceiver_deserializer 
      GENERIC MAP (
        g_data_w         => g_data_w,
        g_tr_clk_period  => c_tr_clk_period
      )
      PORT MAP (
        tb_end        => sim_end,
        
        tr_clk        => tr_clk,  
        tr_rst        => tr_rst,
         
        rx_out_data   => rx_sosi_arr(i).data(g_data_w-1 DOWNTO 0),
        rx_out_ctrl   => rx_out_ctrl(i),
        rx_out_sop    => rx_out_sop(i),
        rx_out_eop    => rx_out_eop(i),
    
        rx_serial_in  => rx_datain(i)
      );
    END GENERATE;

    rx_sosi_arr(i).valid <= andv(rx_out_ctrl(i));
    rx_sosi_arr(i).sop   <= andv(rx_out_sop(i));
    rx_sosi_arr(i).eop   <= andv(rx_out_eop(i));

  END GENERATE;

END str;

