-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, technology_lib, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE technology_lib.technology_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.tech_tse_pkg.ALL;


ENTITY tech_tse IS
  GENERIC (
    g_technology   : NATURAL := c_tech_select_default;
    g_ETH_PHY      : STRING  := "LVDS"; -- "LVDS" (default): uses LVDS IOs for ctrl_unb_common, "XCVR": uses tranceiver PHY
    g_sim          : BOOLEAN := FALSE;
    g_sim_level    : NATURAL := 0;     -- 0 = use IP model (equivalent to g_sim = FALSE); 1 = use fast serdes model;
    g_sim_tx       : BOOLEAN := TRUE;
    g_sim_rx       : BOOLEAN := TRUE
  );
  PORT (
    -- Clocks and reset
    mm_rst         : IN  STD_LOGIC;
    mm_clk         : IN  STD_LOGIC;  -- MM
    eth_clk        : IN  STD_LOGIC;  -- 125 MHz
    tx_snk_clk     : IN  STD_LOGIC;  -- DP
    rx_src_clk     : IN  STD_LOGIC;  -- DP
    
    -- Calibration & reconfig clock
    cal_rec_clk    : IN  STD_LOGIC := '0';
    
    -- Memory Mapped Slave
    mm_sla_in      : IN  t_mem_mosi;
    mm_sla_out     : OUT t_mem_miso;
    
    -- MAC transmit interface
    -- . ST sink
    tx_snk_in      : IN  t_dp_sosi;
    tx_snk_out     : OUT t_dp_siso;
    -- . MAC specific
    tx_mac_in      : IN  t_tech_tse_tx_mac;
    tx_mac_out     : OUT t_tech_tse_tx_mac;
    
    -- MAC receive interface
    -- . ST Source
    rx_src_in      : IN  t_dp_siso;
    rx_src_out     : OUT t_dp_sosi;
    -- . MAC specific
    rx_mac_out     : OUT t_tech_tse_rx_mac;

    -- PHY interface
    eth_txp        : OUT STD_LOGIC;
    eth_rxp        : IN  STD_LOGIC;

    tse_led        : OUT t_tech_tse_led
  );
END tech_tse;

ARCHITECTURE str OF tech_tse IS

  CONSTANT c_use_technology : BOOLEAN := g_sim = FALSE OR g_sim_level = 0;
  CONSTANT c_use_sim_model  : BOOLEAN := NOT c_use_technology;
  
  COMPONENT sim_tse IS 
  GENERIC(
    g_tx         : BOOLEAN;
    g_tx_crc     : BOOLEAN := TRUE;  -- model append CRC by TSE MAC, CRC value = 0
    g_rx         : BOOLEAN
  );      
  PORT(
    -- Clocks and reset
    mm_rst         : IN  STD_LOGIC;  -- unused
    mm_clk         : IN  STD_LOGIC;  -- unused
    eth_clk        : IN  STD_LOGIC;  -- 125 MHz
    tx_snk_clk     : IN  STD_LOGIC;  -- DP
    rx_src_clk     : IN  STD_LOGIC;  -- DP
       
    -- Memory Mapped Slave
    mm_sla_in      : IN  t_mem_mosi;
    mm_sla_out     : OUT t_mem_miso;
    
    -- MAC transmit interface
    -- . ST sink
    tx_snk_in      : IN  t_dp_sosi;
    tx_snk_out     : OUT t_dp_siso;
    -- . MAC specific
    tx_mac_in      : IN  t_tech_tse_tx_mac;
    tx_mac_out     : OUT t_tech_tse_tx_mac;
    
    -- MAC receive interface
    -- . ST Source
    rx_src_in      : IN  t_dp_siso;
    rx_src_out     : OUT t_dp_sosi;
    -- . MAC specific
    rx_mac_out     : OUT t_tech_tse_rx_mac;

    -- PHY interface
    eth_txp        : OUT STD_LOGIC;
    eth_rxp        : IN  STD_LOGIC;

    tse_led        : OUT t_tech_tse_led
  );
END COMPONENT;

BEGIN

  gen_ip_stratixiv : IF c_use_technology=TRUE AND g_technology=c_tech_stratixiv GENERATE
    u0 : ENTITY work.tech_tse_stratixiv
    GENERIC MAP (g_ETH_PHY)
    PORT MAP (mm_rst, mm_clk, eth_clk, tx_snk_clk, rx_src_clk,
              cal_rec_clk,
              mm_sla_in, mm_sla_out,
              tx_snk_in, tx_snk_out,
              tx_mac_in, tx_mac_out,
              rx_src_in, rx_src_out,
              rx_mac_out,
              eth_txp, eth_rxp,
              tse_led);
  END GENERATE;
  
  gen_ip_arria10 : IF c_use_technology=TRUE AND g_technology=c_tech_arria10 GENERATE
    u0 : ENTITY work.tech_tse_arria10
    GENERIC MAP (g_ETH_PHY)
    PORT MAP (mm_rst, mm_clk, eth_clk, tx_snk_clk, rx_src_clk,
              mm_sla_in, mm_sla_out,
              tx_snk_in, tx_snk_out,
              tx_mac_in, tx_mac_out,
              rx_src_in, rx_src_out,
              rx_mac_out,
              eth_txp, eth_rxp,
              tse_led);
  END GENERATE;

  gen_ip_arria10_e3sge3 : IF c_use_technology=TRUE AND g_technology=c_tech_arria10_e3sge3 GENERATE
    u0 : ENTITY work.tech_tse_arria10_e3sge3
    GENERIC MAP (g_ETH_PHY)
    PORT MAP (mm_rst, mm_clk, eth_clk, tx_snk_clk, rx_src_clk,
              mm_sla_in, mm_sla_out,
              tx_snk_in, tx_snk_out,
              tx_mac_in, tx_mac_out,
              rx_src_in, rx_src_out,
              rx_mac_out,
              eth_txp, eth_rxp,
              tse_led);
  END GENERATE;

  gen_ip_arria10_e1sg : IF c_use_technology=TRUE AND g_technology=c_tech_arria10_e1sg GENERATE
    u0 : ENTITY work.tech_tse_arria10_e1sg
    GENERIC MAP (g_ETH_PHY)
    PORT MAP (mm_rst, mm_clk, eth_clk, tx_snk_clk, rx_src_clk,
              mm_sla_in, mm_sla_out,
              tx_snk_in, tx_snk_out,
              tx_mac_in, tx_mac_out,
              rx_src_in, rx_src_out,
              rx_mac_out,
              eth_txp, eth_rxp,
              tse_led);
  END GENERATE;

  gen_ip_arria10_e2sg : IF c_use_technology=TRUE AND g_technology=c_tech_arria10_e2sg GENERATE
    u0 : ENTITY work.tech_tse_arria10_e2sg
    GENERIC MAP (g_ETH_PHY)
    PORT MAP (mm_rst, mm_clk, eth_clk, tx_snk_clk, rx_src_clk,
              mm_sla_in, mm_sla_out,
              tx_snk_in, tx_snk_out,
              tx_mac_in, tx_mac_out,
              rx_src_in, rx_src_out,
              rx_mac_out,
              eth_txp, eth_rxp,
              tse_led);
  END GENERATE;

  gen_sim_tse : IF c_use_sim_model=TRUE GENERATE
    u_sim_tse : sim_tse
    GENERIC MAP (g_sim_tx, TRUE, g_sim_rx)
    PORT MAP (mm_rst, mm_clk, eth_clk, tx_snk_clk, rx_src_clk,
              mm_sla_in, mm_sla_out,
              tx_snk_in, tx_snk_out,
              tx_mac_in, tx_mac_out,
              rx_src_in, rx_src_out,
              rx_mac_out,
              eth_txp, eth_rxp,
              tse_led);
  END GENERATE;
  
END ARCHITECTURE;
