-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;

PACKAGE tech_tse_pkg IS

  CONSTANT c_tech_tse_reg_addr_w           : NATURAL := 8;         -- = max 256 MAC registers
  CONSTANT c_tech_tse_byte_addr_w          : NATURAL := c_tech_tse_reg_addr_w + 2;
  CONSTANT c_tech_tse_byte_addr_pcs_offset : NATURAL := 16#200#;   -- table 4.8, 4.9 in ug_ethernet.pdf
  CONSTANT c_tech_tse_data_w               : NATURAL := c_word_w;  -- = 32

  CONSTANT c_tech_tse_symbol_w             : NATURAL := c_byte_w;                       -- = 8
  CONSTANT c_tech_tse_symbol_max           : NATURAL := 2**c_tech_tse_symbol_w-1;            -- = 255
  CONSTANT c_tech_tse_symbols_per_beat     : NATURAL := c_tech_tse_data_w / c_tech_tse_symbol_w;  -- = 4

  CONSTANT c_tech_tse_pcs_reg_addr_w       : NATURAL := 5;         -- = max 32 PCS registers
  CONSTANT c_tech_tse_pcs_halfword_addr_w  : NATURAL := c_tech_tse_pcs_reg_addr_w + 1;  -- table 4.17 in ug_ethernet.pdf
  CONSTANT c_tech_tse_pcs_byte_addr_w      : NATURAL := c_tech_tse_pcs_reg_addr_w + 2;
  CONSTANT c_tech_tse_pcs_data_w           : NATURAL := c_halfword_w;  -- = 16;

  CONSTANT c_tech_tse_empty_w              : NATURAL := 2;
  CONSTANT c_tech_tse_tx_error_w           : NATURAL := 1;
  CONSTANT c_tech_tse_rx_error_w           : NATURAL := 6;
  CONSTANT c_tech_tse_error_w              : NATURAL := largest(c_tech_tse_tx_error_w, c_tech_tse_rx_error_w);
  CONSTANT c_tech_tse_err_stat_w           : NATURAL := 18;
  CONSTANT c_tech_tse_frm_type_w           : NATURAL := 4;
  
  CONSTANT c_tech_tse_rx_ready_latency     : NATURAL := 2;    -- 2 = default when FIFO is used
  CONSTANT c_tech_tse_tx_ready_latency     : NATURAL := 1;    -- c_tech_tse_tx_ready_latency + 3 = TX_ALMOST_FULL

  CONSTANT c_tech_tse_tx_fifo_depth        : NATURAL := 256;  -- nof words for Tx FIFO
  CONSTANT c_tech_tse_rx_fifo_depth        : NATURAL := 256;  -- nof words for Rx FIFO

  TYPE t_tech_tse_tx_mac IS RECORD
    -- Tx MAC inputs
    crc_fwd  : STD_LOGIC;
    -- Tx MAC outputs
    septy    : STD_LOGIC;
    a_full   : STD_LOGIC;
    a_empty  : STD_LOGIC;
    uflow    : STD_LOGIC;
  END RECORD;

  TYPE t_tech_tse_rx_mac IS RECORD
    -- Rx MAC inputs
    -- Rx MAC outputs
    ethertype: STD_LOGIC_VECTOR(c_tech_tse_err_stat_w-1 DOWNTO 0);
    frm_type : STD_LOGIC_VECTOR(c_tech_tse_frm_type_w-1 DOWNTO 0);
    dsav     : STD_LOGIC;
    a_full   : STD_LOGIC;
    a_empty  : STD_LOGIC;
  END RECORD;

  TYPE t_tech_tse_led IS RECORD
    an       : STD_LOGIC;
    link     : STD_LOGIC;
    disp_err : STD_LOGIC;
    char_err : STD_LOGIC;
    crs      : STD_LOGIC;
    col      : STD_LOGIC;
  END RECORD;
  
  FUNCTION func_tech_tse_map_pcs_addr(pcs_addr : NATURAL) RETURN NATURAL;

END tech_tse_pkg;


PACKAGE BODY tech_tse_pkg IS

FUNCTION func_tech_tse_map_pcs_addr(pcs_addr : NATURAL) RETURN NATURAL IS
BEGIN
  RETURN pcs_addr * 2 + c_tech_tse_byte_addr_pcs_offset;
END func_tech_tse_map_pcs_addr;

END tech_tse_pkg;
